/*
 * Copyright (C) 2007 Oracle.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 021110-1307, USA.
 */
#define _XOPEN_SOURCE 500
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <uuid/uuid.h>

#include "kerncompat.h"
#include "list.h"
#include "radix-tree.h"
#include "ctree.h"
#include "extent-cache.h"
#include "disk-io.h"
#include "volumes.h"
#include "transaction.h"
#include "crc32c.h"
#include "utils.h"
#include "version.h"

static struct recover_control *init_recover_control()
{
	struct recover_control *rc;

	rc = malloc(sizeof(struct recover_control));
	if (!rc)
		return NULL;

	memset(rc, 0, sizeof(struct recover_control));

	INIT_LIST_HEAD(&rc->sys_chunk_list);
	INIT_LIST_HEAD(&rc->non_sys_chunk_list);

	return rc;
}

static int free_recover_control(struct recover_control *rc)
{
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	if (!rc)
		return -1;

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		if (cc->chunk)
			free(cc->chunk);
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		if (cc->chunk)
			free(cc->chunk);
	}

	free(rc);

	return 0;
}

static void print_device(struct recover_control *rc)
{
	struct list_head *cur;
	struct list_head *head;
	struct btrfs_device *dev;
	char str[37];

	printf("device list:\n");
	head = &rc->fs_devices->devices;
	list_for_each(cur, head) {
		dev = list_entry(cur, struct btrfs_device, dev_list);
		uuid_unparse(dev->uuid, str);
		printf("devid:%llu, name:%s, uuid:%s\n",
				dev->devid, dev->name, str);
	}
	printf("\n");
}

static int cache_chunk_is_empty(struct recover_control *rc)
{
	if (list_empty(&rc->sys_chunk_list))
		return 1;
	else
		return 0;
}

static int print_cache_chunk(struct recover_control *rc)
{
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	if (rc->silent)
		return 0;

	printf("chunk list:\n");
	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		if (cc->chunk)
			printf("system chunk [%llu, %llu]\n",
					cc->offset, cc->chunk->length);
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		if (cc->chunk)
			printf("normal chunk [%llu, %llu]\n",
					cc->offset, cc->chunk->length);
	}
	printf("\n");

	return 0;
}

static struct btrfs_chunk *create_cache_chunk(struct recover_control *rc,
				struct btrfs_stripe_header *header)
{
	u64 flag;
	u32 num_stripes;
	u64 len;
	struct cache_chunk *cc;
	struct btrfs_chunk *chunk;

	cc = malloc(sizeof(struct cache_chunk));
	if (!cc)
		return NULL;

	cc->offset = header->chunk_offset;
	num_stripes = header->num_stripes;

	len = btrfs_chunk_item_size(num_stripes);
	chunk = malloc(len);
	if (!chunk) {
		free(cc);
		return NULL;
	}

	memset(chunk, 0, len);
	chunk->length = header->chunk_size;
	chunk->owner = header->owner;
	chunk->stripe_len = header->stripe_len;
	chunk->type = header->type;
	chunk->io_align = header->io_align;
	chunk->io_width = header->io_width;
	chunk->sector_size = header->sector_size;
	chunk->num_stripes = num_stripes;
	chunk->sub_stripes = header->sub_stripes;

	cc->chunk = chunk;

	flag = header->type;
	if (flag & BTRFS_BLOCK_GROUP_SYSTEM)
		list_add(&cc->chunk_list, &rc->sys_chunk_list);
	else
		list_add(&cc->chunk_list, &rc->non_sys_chunk_list);

	return chunk;
}

static struct btrfs_chunk *find_cache_chunk(struct recover_control *rc,
				struct btrfs_stripe_header *header)
{
	u64 offset;
	u64 flag;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	offset = header->chunk_offset;
	flag = header->type;

	if (flag & BTRFS_BLOCK_GROUP_SYSTEM)
		list = &rc->sys_chunk_list;
	else
		list = &rc->non_sys_chunk_list;

	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		if (cc->offset == offset)
			return cc->chunk;
	}

	return NULL;
}

static int update_cache_chunk(struct recover_control *rc,
				struct btrfs_stripe_header *header)
{
	u64 offset;
	u64 flag;
	struct btrfs_stripe *stripes;
	struct btrfs_stripe *stripe;
	struct btrfs_chunk *chunk = NULL;

	if (!rc || !header)
		return -EFAULT;

	offset = header->chunk_offset;
	flag = header->type;

	chunk = find_cache_chunk(rc, header);
	if (chunk) {
		if (header->type != chunk->type ||
			header->chunk_size != chunk->length ||
			header->num_stripes != chunk->num_stripes ||
			header->sub_stripes != chunk->sub_stripes ||
			header->stripe_index >= chunk->num_stripes)
			return -EIO;
	} else {
		if (header->stripe_index < header->num_stripes)
			chunk = create_cache_chunk(rc, header);
		if (!chunk)
			return -ENOMEM;
	}

	stripes = &chunk->stripe;
	stripe = stripes + header->stripe_index;
	btrfs_set_stack_stripe_devid(stripe, header->devid);
	btrfs_set_stack_stripe_offset(stripe, header->dev_offset);
	memcpy(stripe->dev_uuid, header->uuid, BTRFS_UUID_SIZE);
	return 0;
}

#define MAX_STRIPE_ERR_TYPE 8
static char *stripe_header_err[MAX_STRIPE_ERR_TYPE] = {
	"CHECK STRIPE HEADER:disk address invalid",
	"CHECK STRIPE HEADER:stripes number zero",
	"CHECK STRIPE HEADER:UUID unmatch",
	"CHECK STRIPE HEADER:FSID unmatch",
	"CHECK STRIPE HEADER:crc unmatch"
};

static int check_stripe_header(struct recover_control *rc,
				struct btrfs_stripe_header *header,
				struct btrfs_super_block *sb,
				u64 bytenr)
{
	int err = 0;
	u32 crc_s;
	u32 crc_r;

	if (header->dev_offset != bytenr) {
		err = 1;
		goto err;
	}

	if (header->num_stripes == 0) {
		err = 2;
		goto err;
	}

	if (memcmp(header->uuid, sb->dev_item.uuid, BTRFS_UUID_SIZE)) {
		err = 3;
		goto err;
	}

	if (memcmp(header->fsid, sb->fsid, BTRFS_FSID_SIZE)) {
		err = 4;
		goto err;
	}

	crc_s = header->crc;
	header->crc = 0;
	crc_r = crc32c(0, (unsigned char *)header, sizeof(*header));
	if (crc_s != crc_r)
		err = 5;

err:
	if (err && !rc->silent)
		fprintf(stderr, "%s\n", stripe_header_err[err - 1]);

	return err;
}

#define MAX_CHUNK_ERR_TYPE 8
static char *chunk_err[MAX_CHUNK_ERR_TYPE] = {
	"CHECK CHUNK:stripe_len is invalid",
	"CHECK CHUNK:chunk type is invalid",
	"CHECK CHUNK:some stripe invalid"
};

static int check_one_cache_chunk(struct recover_control *rc,
				struct cache_chunk *cc)
{
	int err = 0;
	struct btrfs_stripe *stripes;
	struct btrfs_stripe *stripe;
	struct btrfs_chunk *chunk;
	int num_stripes;
	int i;

	chunk = cc->chunk;
	num_stripes = chunk->num_stripes;
	stripes = &chunk->stripe;

	if (chunk->stripe_len > 64 * 1024) {
		err = 1;
		goto err;
	}

	if (chunk->type > BTRFS_BLOCK_GROUP_RAID10) {
		err = 2;
		goto err;
	}

	for (i = 0; i < num_stripes; i++) {
		stripe = stripes + i;
		if (stripe->devid == 0) {
			err = 3;
			goto err;
		}
	}

err:
	if (err && !rc->silent)
		fprintf(stderr, "%s\n", chunk_err[err - 1]);
	return err;
}

static int check_cache_chunks(struct recover_control *rc)
{
	int ret;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = check_one_cache_chunk(rc, cc);
		if (ret)
			return ret;
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = check_one_cache_chunk(rc, cc);
		if (ret)
			return ret;
	}

	return 0;
}

static int match_one_cache_chunk(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root,
				struct cache_chunk *cc)
{
	int ret = 0;
	int i;
	int slot;
	u64 offset;
	u16 num_stripes;
	struct btrfs_path *path;
	struct btrfs_key key;
	struct btrfs_root *dev_root;
	struct btrfs_chunk *chunk;
	struct btrfs_stripe *stripes;
	struct btrfs_stripe *stripe;
	struct btrfs_dev_extent *dev_extent;
	struct extent_buffer *l;

	dev_root = root->fs_info->dev_root;
	offset = cc->offset;
	chunk = cc->chunk;
	stripes = &chunk->stripe;

	num_stripes = chunk->num_stripes;
	for (i = 0; i < num_stripes; i++) {
		stripe = stripes + i;
		key.objectid = stripe->devid;
		key.offset = stripe->offset;
		key.type = BTRFS_DEV_EXTENT_KEY;

		path = btrfs_alloc_path();
		if (!path)
			return -ENOMEM;
		btrfs_init_path(path);
		ret = btrfs_search_slot(trans, dev_root, &key, path, 0, 0);
		if (ret) {
			btrfs_release_path(root, path);
			return ret;
		}

		l = path->nodes[0];
		slot = path->slots[0];
		dev_extent = btrfs_item_ptr(l, slot, struct btrfs_dev_extent);
		if (offset != btrfs_dev_extent_chunk_offset(l, dev_extent)) {
			if (!rc->silent)
				printf("device tree unmatch with chunks\n"
						"dev_extent[%llu, %llu], chunk[%llu, %llu]\n",
						btrfs_dev_extent_chunk_offset(l, dev_extent),
						btrfs_dev_extent_length(l, dev_extent),
						offset, chunk->length);

			ret = -1;
			btrfs_release_path(root, path);
			return ret;
		}
		btrfs_release_path(root, path);
	}

	return ret;
}

static int match_cache_chunks(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = match_one_cache_chunk(trans, rc, root, cc);
		if (ret)
			return ret;
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = match_one_cache_chunk(trans, rc, root, cc);
		if (ret)
			return ret;
	}

	return 0;
}

static u64 stripe_bytes_by_type(struct btrfs_stripe_header *header)
{
	u64 type;
	int num_stripes;
	int sub_stripes;
	u64 chunk_bytes;

	chunk_bytes = header->chunk_size;
	num_stripes = header->num_stripes;
	sub_stripes = header->sub_stripes;
	type = header->type;

	if (type & (BTRFS_BLOCK_GROUP_RAID1 | BTRFS_BLOCK_GROUP_DUP))
		return chunk_bytes;
	else if (type & BTRFS_BLOCK_GROUP_RAID10)
		return chunk_bytes * sub_stripes / num_stripes;
	else
		return chunk_bytes / num_stripes;
}

static int scan_one_device_cache_chunk(struct recover_control *rc,
				int fd)
{
	int ret = 0;
	char *buf;
	u64 bytenr;
	u64 sectorsize;
	struct btrfs_stripe_header *header;
	struct btrfs_super_block *sb;

	sectorsize = rc->sectorsize;
	buf = malloc(sectorsize);
	if (!buf)
		return -ENOMEM;

	sb = malloc(sizeof(struct btrfs_super_block));
	if (!sb) {
		free(buf);
		return -ENOMEM;
	}

	ret = btrfs_read_dev_super(fd, sb, BTRFS_SUPER_INFO_OFFSET);
	if (ret) {
		ret = -ENOENT;
		goto err;
	}

	bytenr = 0;
	while (1) {
		memset(buf, 0, sectorsize);
		if (pread64(fd, buf, sectorsize, bytenr) < sectorsize)
			break;

		header = (struct btrfs_stripe_header *)buf;
		if (header->tag == BTRFS_STRIPE_HEADER_TAG &&
			!check_stripe_header(rc, header, sb, bytenr)) {
			ret = update_cache_chunk(rc, header);
			if (ret)
				goto err;

			bytenr += stripe_bytes_by_type(header);
		} else {
			bytenr += sectorsize;
		}
	}

err:
	free(sb);
	free(buf);
	return ret;
}

static int scan_devices_cache_chunk(struct recover_control *rc)
{
	int ret = 0;
	int fd;

	struct list_head *cur;
	struct list_head *head;
	struct btrfs_device *dev;

	if (!rc)
		return -EFAULT;

	head = &rc->fs_devices->devices;
	list_for_each(cur, head) {
		dev = list_entry(cur, struct btrfs_device, dev_list);
		fd = open(dev->name, O_RDONLY, 0600);
		if (!fd)
			return -ENOENT;

		ret = scan_one_device_cache_chunk(rc, fd);

		close(fd);

		if (ret)
			return ret;
	}

	return ret;
}

static int read_one_cache_chunk(struct btrfs_root *root,
				struct cache_chunk *cc)
{
	int ret = 0;
	int i;
	u64 devid;
	u8 uuid[BTRFS_UUID_SIZE];
	int num_stripes;
	struct btrfs_mapping_tree *map_tree;
	struct map_lookup *map;
	struct btrfs_stripe *stripes;
	struct btrfs_stripe *stripe;
	struct btrfs_chunk *chunk;

	map_tree = &root->fs_info->mapping_tree;
	chunk = cc->chunk;
	num_stripes = chunk->num_stripes;
	map = malloc(map_lookup_size(num_stripes));
	if (!map)
		return -ENOMEM;

	map->ce.start = cc->offset;
	map->ce.size = chunk->length;
	map->num_stripes = num_stripes;
	map->io_width = chunk->io_width;
	map->io_align = chunk->io_align;
	map->sector_size = chunk->sector_size;
	map->stripe_len = chunk->stripe_len;
	map->type = chunk->type;
	map->sub_stripes = chunk->sub_stripes;

	stripes = &chunk->stripe;
	for (i = 0; i < num_stripes; i++) {
		stripe = stripes + i;
		devid = stripe->devid;
		memcpy(uuid, stripe->dev_uuid, BTRFS_UUID_SIZE);
		map->stripes[i].physical = stripe->offset;
		map->stripes[i].dev = btrfs_find_device(root, devid, uuid, NULL);

		if (!map->stripes[i].dev) {
			kfree(map);
			return -EIO;
		}
	}

	ret = insert_existing_cache_extent(&map_tree->cache_tree, &map->ce);
	return ret;
}

static int read_cache_chunks(struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = read_one_cache_chunk(root, cc);
		if (ret)
			return ret;
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = read_one_cache_chunk(root, cc);
		if (ret)
			return ret;
	}

	return 0;
}

static int __remove_chunk_extent_item(struct btrfs_trans_handle *trans,
				struct btrfs_root *root,
				u64 start, u64 offset)
{
	int ret;
	struct btrfs_key key;
	struct btrfs_path *path;

	root = root->fs_info->extent_root;
	key.objectid = start;
	key.offset = offset;
	key.type = BTRFS_EXTENT_ITEM_KEY;

	path = btrfs_alloc_path();
	if (!path)
		return -ENOMEM;

	ret = btrfs_search_slot(trans, root, &key, path, 0, 1);
	if (ret < 0)
		goto err;
	else if (ret > 0) {
		ret = 0;
		goto err;
	} else
		ret = btrfs_del_item(trans, root, path);

err:
	btrfs_free_path(path);
	return ret;
}

static int remove_chunk_extent_item(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret = 0;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;
	u64 start;
	u64 end;
	u64 sectorsize;

	sectorsize = rc->sectorsize;
	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		start = cc->offset;
		end = cc->offset + cc->chunk->length;

		while (start < end) {
			ret = __remove_chunk_extent_item(trans, root,
						start, sectorsize);
			if (ret)
				return ret;
			start += sectorsize;
		}
	}

	return 0;
}

static int reset_block_group(struct btrfs_trans_handle *trans,
				struct btrfs_root *root,
				u64 bytenr, u64 num_bytes)
{
	int ret = 0;
	struct btrfs_block_group_cache *cache;
	struct btrfs_fs_info *info;
	u64 byte_in_group;
	u64 total;
	u64 start;
	u64 end;

	info = root->fs_info;
	total = num_bytes;
	while (total) {
		cache = btrfs_lookup_block_group(info, bytenr);
		if (!cache)
			return -1;

		start = cache->key.objectid;
		end = start + cache->key.offset - 1;
		set_extent_bits(&info->block_group_cache, start, end,
						EXTENT_DIRTY, GFP_NOFS);

		byte_in_group = bytenr - cache->key.objectid;
		num_bytes =  min(total, cache->key.offset - byte_in_group);

		set_extent_dirty(&info->free_space_cache, bytenr,
				bytenr + num_bytes - 1, GFP_NOFS);

		btrfs_set_block_group_used(&cache->item, 0);
		total -= num_bytes;
		bytenr += num_bytes;
	}

	return ret;
}

static int clean_sys_block_group_info(struct btrfs_trans_handle *trans,
					struct recover_control *rc,
					struct btrfs_root *root)
{
	int ret = 0;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		ret = reset_block_group(trans, root, cc->offset,
				cc->chunk->length);
		if (ret)
			return ret;
	}
	return ret;
}

static int __reset_chunk_root(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret;
	u64 min_devid;
	struct list_head *head;
	struct list_head *cur;
	struct btrfs_super_block *super_copy;
	struct btrfs_device *dev;
	struct extent_buffer *cow;
	struct btrfs_disk_key disk_key;

	ret = 0;
	min_devid = 1;
	head = &rc->fs_devices->devices;
	list_for_each(cur, head) {
		dev = list_entry(cur, struct btrfs_device, dev_list);
		if (min_devid > dev->devid)
			min_devid = dev->devid;
	}
	disk_key.objectid = BTRFS_DEV_ITEMS_OBJECTID;
	disk_key.type = BTRFS_DEV_ITEM_KEY;
	disk_key.offset = min_devid;

	cow = btrfs_alloc_free_block(trans, root, root->sectorsize,
			BTRFS_CHUNK_TREE_OBJECTID, &disk_key, 0, 0, 0);

	btrfs_set_header_bytenr(cow, cow->start);
	btrfs_set_header_generation(cow, trans->transid);
	btrfs_set_header_nritems(cow, 0);
	btrfs_set_header_level(cow, 0);
	btrfs_set_header_backref_rev(cow, BTRFS_MIXED_BACKREF_REV);
	btrfs_set_header_owner(cow, BTRFS_CHUNK_TREE_OBJECTID);
	write_extent_buffer(cow, root->fs_info->fsid,
			(unsigned long)btrfs_header_fsid(cow),
			BTRFS_FSID_SIZE);

	write_extent_buffer(cow, root->fs_info->chunk_tree_uuid,
			(unsigned long)btrfs_header_chunk_tree_uuid(cow),
			BTRFS_UUID_SIZE);

	root->node = cow;
	btrfs_mark_buffer_dirty(cow);

	super_copy = &root->fs_info->super_copy;
	btrfs_set_super_chunk_root(super_copy, cow->start);
	btrfs_set_super_chunk_root_generation(super_copy, trans->transid);
	btrfs_set_super_chunk_root_level(super_copy, 0);

	return ret;
}

static int __rebuild_device_items(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret = 0;
	struct list_head *cur;
	struct list_head *head;
	struct btrfs_device *dev;
	struct btrfs_key key;
	struct btrfs_dev_item *dev_item;

	head = &rc->fs_devices->devices;
	list_for_each(cur, head) {
		dev = list_entry(cur, struct btrfs_device, dev_list);

		key.objectid = BTRFS_DEV_ITEMS_OBJECTID;
		key.type = BTRFS_DEV_ITEM_KEY;
		key.offset = dev->devid;

		dev_item = malloc(sizeof(struct btrfs_dev_item));
		if (!dev_item)
			return -ENOMEM;

		btrfs_set_stack_device_generation(dev_item, 0);
		btrfs_set_stack_device_type(dev_item, dev->type);
		btrfs_set_stack_device_id(dev_item, dev->devid);
		btrfs_set_stack_device_total_bytes(dev_item, dev->total_bytes);
		btrfs_set_stack_device_bytes_used(dev_item, dev->bytes_used);
		btrfs_set_stack_device_io_align(dev_item, dev->io_align);
		btrfs_set_stack_device_io_width(dev_item, dev->io_width);
		btrfs_set_stack_device_sector_size(dev_item, dev->sector_size);
		memcpy(dev_item->uuid, dev->uuid, BTRFS_UUID_SIZE);
		memcpy(dev_item->fsid, dev->fs_devices->fsid, BTRFS_UUID_SIZE);

		ret = btrfs_insert_item(trans, root, &key,
				dev_item, sizeof(*dev_item));
	}

	return ret;
}

static int __rebuild_chunk_items(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret = 0;
	struct btrfs_key key;
	struct btrfs_chunk *chunk;
	struct btrfs_root *chunk_root;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	chunk_root = root->fs_info->chunk_root;
	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		chunk = cc->chunk;
		key.objectid = BTRFS_FIRST_CHUNK_TREE_OBJECTID;
		key.type = BTRFS_CHUNK_ITEM_KEY;
		key.offset = cc->offset;

		ret = btrfs_insert_item(trans, chunk_root, &key, chunk,
				btrfs_chunk_item_size(chunk->num_stripes));
		if (ret)
			return ret;
	}

	list = &rc->non_sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		chunk = cc->chunk;
		key.objectid = BTRFS_FIRST_CHUNK_TREE_OBJECTID;
		key.type = BTRFS_CHUNK_ITEM_KEY;
		key.offset = cc->offset;

		ret = btrfs_insert_item(trans, chunk_root, &key, chunk,
						btrfs_chunk_item_size(chunk->num_stripes));
		if (ret)
			return ret;
	}

	return ret;
}

static int rebuild_chunk_tree(struct btrfs_trans_handle *trans,
				struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret = 0;

	root = root->fs_info->chunk_root;

	ret = __reset_chunk_root(trans, rc, root);
	if (ret)
		return ret;

	/* insert all the devs */
	ret = __rebuild_device_items(trans, rc, root);
	if (ret)
		return ret;

	/* insert all the chunks */
	ret = __rebuild_chunk_items(trans, rc, root);

	return ret;
}

static int rebuild_sys_array(struct recover_control *rc,
				struct btrfs_root *root)
{
	int ret;
	u32 num_stripes;
	struct btrfs_chunk *chunk;
	struct btrfs_key key;
	struct list_head *list;
	struct list_head *next;
	struct cache_chunk *cc;

	btrfs_set_super_sys_array_size(&root->fs_info->super_copy, 0);

	list = &rc->sys_chunk_list;
	list_for_each(next, list) {
		cc = list_entry(next, struct cache_chunk, chunk_list);
		chunk = cc->chunk;
		num_stripes = chunk->num_stripes;

		key.objectid = BTRFS_FIRST_CHUNK_TREE_OBJECTID;
		key.type = BTRFS_CHUNK_ITEM_KEY;
		key.offset = cc->offset;

		ret = btrfs_add_system_chunk(NULL, root, &key, chunk,
					btrfs_chunk_item_size(num_stripes));
		if (ret)
			return ret;
	}

	return 0;
}

static struct btrfs_root *open_ctree_with_broken_chunk(
				struct recover_control *rc,
				const char *path,
				int writes)
{
	int ret;
	int fd;
	u32 sectorsize;
	u32 nodesize;
	u32 leafsize;
	u32 blocksize;
	u32 stripesize;
	u64 generation;
	u64 sb_bytenr;
	u64 features;
	struct btrfs_key key;
	struct btrfs_root *tree_root = malloc(sizeof(struct btrfs_root));
	struct btrfs_root *extent_root = malloc(sizeof(struct btrfs_root));
	struct btrfs_root *chunk_root = malloc(sizeof(struct btrfs_root));
	struct btrfs_root *dev_root = malloc(sizeof(struct btrfs_root));
	struct btrfs_root *csum_root = malloc(sizeof(struct btrfs_root));
	struct btrfs_fs_info *fs_info = malloc(sizeof(struct btrfs_fs_info));
	struct btrfs_fs_devices *fs_devices = NULL;
	struct btrfs_super_block *disk_super = NULL;

	fd = rc->fd;
	fs_devices = rc->fs_devices;
	sb_bytenr = BTRFS_SUPER_INFO_OFFSET;

	memset(fs_info, 0, sizeof(struct btrfs_fs_info));
	fs_info->rc = rc;
	fs_info->tree_root = tree_root;
	fs_info->extent_root = extent_root;
	fs_info->chunk_root = chunk_root;
	fs_info->dev_root = dev_root;
	fs_info->csum_root = csum_root;

	extent_io_tree_init(&fs_info->extent_cache);
	extent_io_tree_init(&fs_info->free_space_cache);
	extent_io_tree_init(&fs_info->block_group_cache);
	extent_io_tree_init(&fs_info->pinned_extents);
	extent_io_tree_init(&fs_info->pending_del);
	extent_io_tree_init(&fs_info->extent_ins);

	cache_tree_init(&fs_info->fs_root_cache);
	cache_tree_init(&fs_info->mapping_tree.cache_tree);

	mutex_init(&fs_info->fs_mutex);
	fs_info->fs_devices = fs_devices;
	INIT_LIST_HEAD(&fs_info->dirty_cowonly_roots);
	INIT_LIST_HEAD(&fs_info->space_info);

	__setup_root(4096, 4096, 4096, 4096, tree_root,
				fs_info, BTRFS_ROOT_TREE_OBJECTID);

	ret = btrfs_open_devices(fs_devices, O_RDWR);

	fs_info->super_bytenr = sb_bytenr;
	disk_super = &fs_info->super_copy;
	ret = btrfs_read_dev_super(fs_devices->latest_bdev,
				disk_super, sb_bytenr);
	if (ret) {
		printk("No valid btrfs found\n");
		BUG_ON(1);
	}

	memcpy(fs_info->fsid, &disk_super->fsid, BTRFS_FSID_SIZE);

	features = btrfs_super_incompat_flags(disk_super) &
		   ~BTRFS_FEATURE_INCOMPAT_SUPP;
	if (features) {
		printk("couldn't open because of unsupported "
		       "option features (%Lx).\n", features);
		BUG_ON(1);
	}

	features = btrfs_super_incompat_flags(disk_super);
	if (!(features & BTRFS_FEATURE_INCOMPAT_MIXED_BACKREF)) {
		features |= BTRFS_FEATURE_INCOMPAT_MIXED_BACKREF;
		btrfs_set_super_incompat_flags(disk_super, features);
	}

	features = btrfs_super_compat_ro_flags(disk_super) &
		~BTRFS_FEATURE_COMPAT_RO_SUPP;
	if (writes && features) {
		printk("couldn't open RDWR because of unsupported "
		       "option features (%Lx).\n", features);
		BUG_ON(1);
	}

	nodesize = btrfs_super_nodesize(disk_super);
	leafsize = btrfs_super_leafsize(disk_super);
	sectorsize = btrfs_super_sectorsize(disk_super);
	stripesize = btrfs_super_stripesize(disk_super);
	tree_root->nodesize = nodesize;
	tree_root->leafsize = leafsize;
	tree_root->sectorsize = sectorsize;
	tree_root->stripesize = stripesize;

	ret = rebuild_sys_array(rc, tree_root);
	if (ret)
		return NULL;

	/* creating mapping instead of reading chunk */
	ret = read_cache_chunks(rc, tree_root);
	if (ret)
		return NULL;

	blocksize = btrfs_level_size(tree_root,
			btrfs_super_chunk_root_level(disk_super));
	generation = btrfs_super_chunk_root_generation(disk_super);
	__setup_root(nodesize, leafsize, sectorsize, stripesize,
				chunk_root, fs_info, BTRFS_CHUNK_TREE_OBJECTID);

	blocksize = btrfs_level_size(tree_root,
				btrfs_super_root_level(disk_super));
	generation = btrfs_super_generation(disk_super);

	tree_root->node = read_tree_block(tree_root,
				btrfs_super_root(disk_super),
				blocksize, generation);
	if (!tree_root->node)
		return NULL;

	read_extent_buffer(tree_root->node, fs_info->chunk_tree_uuid,
			(unsigned long)btrfs_header_chunk_tree_uuid(tree_root->node),
			BTRFS_UUID_SIZE);

	ret = find_and_setup_root(tree_root, fs_info,
				BTRFS_EXTENT_TREE_OBJECTID, extent_root);
	if (ret)
		return NULL;
	extent_root->track_dirty = 1;

	ret = find_and_setup_root(tree_root, fs_info,
				BTRFS_DEV_TREE_OBJECTID, dev_root);
	if (ret)
		return NULL;
	dev_root->track_dirty = 1;

	ret = find_and_setup_root(tree_root, fs_info,
				BTRFS_CSUM_TREE_OBJECTID, csum_root);
	if (ret)
		return NULL;
	csum_root->track_dirty = 1;

	find_and_setup_log_root(tree_root, fs_info, disk_super);

	fs_info->generation = generation + 1;
	btrfs_read_block_groups(fs_info->tree_root);

	key.objectid = BTRFS_FS_TREE_OBJECTID;
	key.type = BTRFS_ROOT_ITEM_KEY;
	key.offset = (u64)-1;
	fs_info->fs_root = btrfs_read_fs_root(fs_info, &key);

	fs_info->data_alloc_profile = (u64)-1;
	fs_info->metadata_alloc_profile = (u64)-1;
	fs_info->system_alloc_profile = fs_info->metadata_alloc_profile;

	return fs_info->fs_root;
}

static int close_ctree_with_broken_chunk(struct recover_control *rc,
				struct btrfs_root *root)
{
	struct btrfs_fs_info *fs_info;

	if (!rc || !root)
		return -1;

	fs_info = root->fs_info;
	fs_info->rc = NULL;

	btrfs_free_block_groups(fs_info);
	free_fs_roots(fs_info);

	if (fs_info->extent_root->node)
		free_extent_buffer(fs_info->extent_root->node);
	if (fs_info->tree_root->node)
		free_extent_buffer(fs_info->tree_root->node);
	if (fs_info->chunk_root->node)
		free_extent_buffer(fs_info->chunk_root->node);
	if (fs_info->dev_root->node)
		free_extent_buffer(fs_info->dev_root->node);
	if (fs_info->csum_root->node)
		free_extent_buffer(fs_info->csum_root->node);

	if (fs_info->log_root_tree) {
		if (fs_info->log_root_tree->node)
			free_extent_buffer(fs_info->log_root_tree->node);
		free(fs_info->log_root_tree);
	}

	extent_io_tree_cleanup(&fs_info->extent_cache);
	extent_io_tree_cleanup(&fs_info->free_space_cache);
	extent_io_tree_cleanup(&fs_info->block_group_cache);
	extent_io_tree_cleanup(&fs_info->pinned_extents);
	extent_io_tree_cleanup(&fs_info->pending_del);
	extent_io_tree_cleanup(&fs_info->extent_ins);

	free(fs_info->tree_root);
	free(fs_info->extent_root);
	free(fs_info->chunk_root);
	free(fs_info->dev_root);
	free(fs_info->csum_root);
	free(fs_info);

	return 0;
}

static int recover_prepare(struct recover_control *rc,
				char *path, int silent)
{
	int ret;
	int fd;
	u64 total_devs;
	struct btrfs_super_block *sb;
	struct btrfs_fs_devices *fs_devices;

	ret = 0;
	fd = open(path, O_CREAT | O_RDWR, 0600);
	if (fd < 0) {
		fprintf(stderr, "open %s\n error", path);
		return -1;
	}

	rc->fd = fd;
	rc->silent = silent;

	sb = malloc(sizeof(struct btrfs_super_block));
	if (!sb) {
		return -ENOMEM;
		goto fail_close_fd;
	}

	ret = btrfs_read_dev_super(fd, sb, BTRFS_SUPER_INFO_OFFSET);
	if (ret) {
		fprintf(stderr, "read super block error\n");
		free(sb);
		goto fail_free_sb;
	}

	rc->sectorsize = sb->sectorsize;

	/* if seed, the result of scanning below will be partial */
	if (sb->flags & BTRFS_SUPER_FLAG_SEEDING) {
		fprintf(stderr, "this device is seed device\n");
		ret = -1;
		goto fail_free_sb;
	}

	ret = btrfs_scan_one_device(fd, path, &fs_devices,
					&total_devs, BTRFS_SUPER_INFO_OFFSET);
	if (ret)
		goto fail_free_sb;

	if (total_devs != 1) {
		ret = btrfs_scan_for_fsid(fs_devices, total_devs, 1);
		if (ret)
			goto fail_free_sb;
	}

	rc->fs_devices = fs_devices;

	if (!rc->silent)
		print_device(rc);

fail_free_sb:
	free(sb);
fail_close_fd:
	close(fd);
	return ret;
}

static int recover_finish(struct recover_control *rc)
{
	if (rc && rc->fd)
		close(rc->fd);

	free_recover_control(rc);
	return 0;
}

static int btrfs_chunk_tree_check(char *path, int silent)
{
	int ret = 0;
	struct recover_control *rc = NULL;

	rc = init_recover_control();
	if (!rc)
		return -ENOMEM;

	ret = recover_prepare(rc, path, silent);
	if (ret) {
		fprintf(stderr, "recover prepare error\n");
		goto fail_free_rc;
	}

	ret = scan_devices_cache_chunk(rc);
	if (ret) {
		fprintf(stderr, "scan chunk headers error\n");
		goto fail_free_rc;
	}

	ret = check_cache_chunks(rc);
	if (ret) {
		fprintf(stderr, "check chunk error\n");
		goto fail_free_rc;
	}

	if (cache_chunk_is_empty(rc)) {
		ret = -1;
		goto fail_free_rc;
	} else
		print_cache_chunk(rc);

fail_free_rc:
	recover_finish(rc);
	return ret;
}

static int btrfs_chunk_tree_recover(char *path, int silent)
{
	int ret = 0;
	struct btrfs_root *root = NULL;
	struct btrfs_trans_handle *trans;
	struct recover_control *rc = NULL;

	rc = init_recover_control();
	if (!rc)
		return -ENOMEM;

	ret = recover_prepare(rc, path, silent);
	if (ret) {
		fprintf(stderr, "recover prepare error\n");
		goto fail_free_rc;
	}

	ret = scan_devices_cache_chunk(rc);
	if (ret) {
		fprintf(stderr, "scan chunk headers error\n");
		goto fail_free_rc;
	}

	ret = check_cache_chunks(rc);
	if (ret) {
		fprintf(stderr, "check chunk error\n");
		goto fail_free_rc;
	}

	if (cache_chunk_is_empty(rc)) {
		fprintf(stderr, "no chunk searched error\n");
		goto fail_free_rc;
	} else
		print_cache_chunk(rc);

	root = open_ctree_with_broken_chunk(rc, path, O_RDWR);
	if (!root) {
		fprintf(stderr, "open with broken chunk error\n");
		goto fail_close_ctree;
	}

	ret = match_cache_chunks(NULL, rc, root);
	if (ret) {
		fprintf(stderr, "match chunk error\n");
		goto fail_close_ctree;
	}

	trans = btrfs_start_transaction(root, 1);
	ret = remove_chunk_extent_item(trans, rc, root);
	BUG_ON(ret);

	ret = clean_sys_block_group_info(trans, rc, root);
	BUG_ON(ret);

	ret = rebuild_chunk_tree(trans, rc, root);
	BUG_ON(ret);
	btrfs_commit_transaction(trans, root);

fail_close_ctree:
	close_ctree_with_broken_chunk(rc, root);
fail_free_rc:
	recover_finish(rc);
	return ret;
}

static void print_usage(void)
{
	fprintf(stderr, "usage:btrfs-recover-chunk [options] dev\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "\t -c --check stripe header after scan dev\n");
	fprintf(stderr, "\t -s --silent mode\n");
	fprintf(stderr, "%s\n", BTRFS_BUILD_VERSION);
	exit(1);
}

int main(int argc, char *argv[])
{
	int ret = 0;
	int silent = 0;
	int check = 0;
	char *file;

	while (1) {
		int c = getopt(argc, argv, "sc");
		if (c < 0)
			break;
		switch (c) {
		case 's':
			silent  = 1;
			break;
		case 'c':
			check = 1;
			break;
		default:
			print_usage();
		}
	}

	argc = argc - optind;
	if (argc == 0)
		print_usage();

	file = argv[optind];

	ret = check_mounted(file);
	if (ret) {
		fprintf(stderr, "the device is busy\n");
		return ret;
	}

	if (silent)
		printf("slient mode enable\n");

	if (check) {
		ret = btrfs_chunk_tree_check(file, silent);
		if (ret)
			printf("some stripe header invalid\n");
		else
			printf("all stripe headers valid\n");
	} else {
		ret = btrfs_chunk_tree_recover(file, silent);
		if (ret)
			printf("rebuild chunk tree fail\n");
		else
			printf("rebuild chunk tree success\n");
	}

	return 0;
}
